#include <algorithm>
#include <cstring>
#include <set>
#include <msp/core/hash.h>
#include <msp/core/maputils.h>
#include <msp/gl/extensions/arb_fragment_shader.h>
#include <msp/gl/extensions/arb_geometry_shader4.h>
#include <msp/gl/extensions/arb_shader_objects.h>
#include <msp/gl/extensions/arb_uniform_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_shader.h>
#include <msp/gl/extensions/ext_gpu_shader4.h>
#include <msp/gl/extensions/nv_non_square_matrices.h>
#include <msp/io/print.h>
#include <msp/strings/format.h>
#include "buffer.h"
#include "error.h"
#include "misc.h"
#include "module.h"
#include "program.h"
#include "resources.h"
#include "shader.h"
#include "glsl/compiler.h"

using namespace std;

namespace Msp {
namespace GL {

Program::Program()
{
	init();
}

Program::Program(const std::string &source)
{
	init();

	Module mod;
	mod.set_source(source);
	add_stages(mod);

	link();
	module = 0;
}

Program::Program(const string &vert, const string &frag)
{
	init();

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	attach_shader_owned(new VertexShader(vert));
	attach_shader_owned(new FragmentShader(frag));
#pragma GCC diagnostic pop
	link();
}

Program::Program(const Module &mod, const map<string, int> &spec_values)
{
	init();
	add_stages(mod, spec_values);
	link();
}

void Program::init()
{
	static Require _req(ARB_shader_objects);

	id = glCreateProgram();
	module = 0;
	linked = false;
}

Program::~Program()
{
	for(vector<unsigned>::iterator i=shader_ids.begin(); i!=shader_ids.end(); ++i)
		glDeleteShader(*i);
	glDeleteProgram(id);
}

void Program::add_stages(const Module &mod, const map<string, int> &spec_values)
{
	module = &mod;

	SL::Compiler compiler;
	compiler.set_source(module->get_prepared_source(), "<module>");
	compiler.specialize(spec_values);
	compiler.compile(SL::Compiler::PROGRAM);

	vector<SL::Stage::Type> stages = compiler.get_stages();
	for(vector<SL::Stage::Type>::const_iterator i=stages.begin(); i!=stages.end(); ++i)
	{
		GLenum type;
		switch(*i)
		{
		case SL::Stage::VERTEX:
			{ static Require _req(ARB_vertex_shader); }
			type = GL_VERTEX_SHADER;
			break;
		case SL::Stage::GEOMETRY:
			{ static Require _req(ARB_geometry_shader4); }
			type = GL_GEOMETRY_SHADER;
			break;
		case SL::Stage::FRAGMENT:
			{ static Require _req(ARB_fragment_shader); }
			type = GL_FRAGMENT_SHADER;
			break;
		default:
			throw invalid_operation("Program::add_stages");
		}

		unsigned stage_id = glCreateShader(type);
		shader_ids.push_back(stage_id);
		glAttachShader(id, stage_id);

		string stage_src = compiler.get_stage_glsl(*i);
		const char *src_ptr = stage_src.data();
		int src_len = stage_src.size();
		glShaderSource(stage_id, 1, &src_ptr, &src_len);

		if(*i==SL::Stage::VERTEX)
		{
			const map<string, unsigned> &attribs = compiler.get_vertex_attributes();
			for(map<string, unsigned>::const_iterator j=attribs.begin(); j!=attribs.end(); ++j)
				glBindAttribLocation(id, j->second, j->first.c_str());
		}

		if(*i==SL::Stage::FRAGMENT && EXT_gpu_shader4)
		{
			const map<string, unsigned> &frag_outs = compiler.get_fragment_outputs();
			for(map<string, unsigned>::const_iterator j=frag_outs.begin(); j!=frag_outs.end(); ++j)
				glBindFragDataLocation(id, j->second, j->first.c_str());
		}
	}
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void Program::attach_shader(Shader &shader)
{
	unsigned shader_id = shader.steal_id();
	if(!shader_id)
		throw invalid_argument("Program::attach_shader");
	shader_ids.push_back(shader_id);
}

void Program::attach_shader_owned(Shader *shader)
{
	attach_shader(*shader);
	delete shader;
}

void Program::detach_shader(Shader &)
{
}

const vector<Shader *> &Program::get_attached_shaders() const
{
	static vector<Shader *> dummy;
	return dummy;
}

void Program::bind_attribute(unsigned index, const string &name)
{
	static Require _req(ARB_vertex_shader);
	glBindAttribLocation(id, index, name.c_str());
}

void Program::bind_attribute(VertexComponent comp, const string &name)
{
	bind_attribute(get_component_type(comp), name);
}

void Program::bind_fragment_data(unsigned index, const string &name)
{
	static Require _req(EXT_gpu_shader4);
	glBindFragDataLocation(id, index, name.c_str());
}
#pragma GCC diagnostic pop

void Program::link()
{
	for(vector<unsigned>::const_iterator i=shader_ids.begin(); i!=shader_ids.end(); ++i)
	{
		glCompileShader(*i);
		bool compiled = get_shader_i(*i, GL_COMPILE_STATUS);

		GLsizei info_log_len = get_shader_i(*i, GL_INFO_LOG_LENGTH);
		string info_log(info_log_len+1, 0);
		glGetShaderInfoLog(*i, info_log_len+1, &info_log_len, &info_log[0]);
		info_log.erase(info_log_len);
		info_log = module->get_source_map().translate_errors(info_log);

		if(!compiled)
			throw compile_error(info_log);
#ifdef DEBUG
		if(!info_log.empty())
			IO::print("Shader compile info log:\n%s", info_log);
#endif
	}

	uniforms.clear();

	glLinkProgram(id);
	linked = get_program_i(id, GL_LINK_STATUS);

	GLsizei info_log_len = get_program_i(id, GL_INFO_LOG_LENGTH);
	string info_log(info_log_len+1, 0);
	glGetProgramInfoLog(id, info_log_len+1, &info_log_len, &info_log[0]);
	info_log.erase(info_log_len);
	info_log = module->get_source_map().translate_errors(info_log);

	if(!linked)
		throw compile_error(info_log);
#ifdef DEBUG
	if(!info_log.empty())
		IO::print("Program link info log:\n%s", info_log);
#endif

	query_uniforms();
	query_attributes();

	for(UniformMap::const_iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
		require_type(i->second.type);
	for(AttributeMap::const_iterator i=attributes.begin(); i!=attributes.end(); ++i)
		require_type(i->second.type);
}

void Program::require_type(GLenum t)
{
	switch(t)
	{
	case GL_FLOAT_MAT2x3:
	case GL_FLOAT_MAT2x4:
	case GL_FLOAT_MAT3x2:
	case GL_FLOAT_MAT3x4:
	case GL_FLOAT_MAT4x2:
	case GL_FLOAT_MAT4x3:
		{ static Require _req(NV_non_square_matrices); }
		break;
	}
}

void Program::query_uniforms()
{
	unsigned count = get_program_i(id, GL_ACTIVE_UNIFORMS);
	vector<UniformInfo *> uniforms_by_index(count);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len = 0;
		int size;
		GLenum type;
		glGetActiveUniform(id, i, sizeof(name), &len, &size, &type, name);
		if(len && strncmp(name, "gl_", 3))
		{
			/* Some implementations report the first element of a uniform array,
			others report just the name of the array itself. */
			if(len>3 && !strcmp(name+len-3, "[0]"))
				name[len-3] = 0;

			UniformInfo &info = uniforms[name];
			info.block = 0;
			info.name = name;
			info.size = size;
			info.array_stride = 0;
			info.matrix_stride = 0;
			info.type = type;
			uniforms_by_index[i] = &info;
		}
	}

	if(ARB_uniform_buffer_object)
		query_uniform_blocks(uniforms_by_index);

	UniformBlockInfo &default_block = uniform_blocks[string()];
	default_block.data_size = 0;
	default_block.bind_point = -1;

	for(UniformMap::iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
		if(!i->second.block)
		{
			i->second.location = glGetUniformLocation(id, i->second.name.c_str());
			i->second.block = &default_block;
			default_block.uniforms.push_back(&i->second);
		}

	default_block.layout_hash = compute_layout_hash(default_block.uniforms);

	string layout_descriptor;
	for(UniformBlockMap::const_iterator i=uniform_blocks.begin(); i!=uniform_blocks.end(); ++i)
		layout_descriptor += format("%d:%x\n", i->second.bind_point, i->second.layout_hash);
	uniform_layout_hash = hash32(layout_descriptor);
}

void Program::query_uniform_blocks(const vector<UniformInfo *> &uniforms_by_index)
{
	uniform_blocks.clear();

	std::set<unsigned> used_bind_points;
	unsigned count = get_program_i(id, GL_ACTIVE_UNIFORM_BLOCKS);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len;
		glGetActiveUniformBlockName(id, i, sizeof(name), &len, name);
		UniformBlockInfo &info = uniform_blocks[name];
		info.name = name;

		int value;
		glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_DATA_SIZE, &value);
		info.data_size = value;

		glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS, &value);
		vector<int> indices(value);
		glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES, &indices[0]);
		for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
		{
			if(!uniforms_by_index[*j])
				throw logic_error("Program::link");
			info.uniforms.push_back(uniforms_by_index[*j]);
			uniforms_by_index[*j]->block = &info;
		}

		vector<unsigned> indices2(indices.begin(), indices.end());
		vector<int> values(indices.size());
		glGetActiveUniformsiv(id, indices.size(), &indices2[0], GL_UNIFORM_OFFSET, &values[0]);
		for(unsigned j=0; j<indices.size(); ++j)
			uniforms_by_index[indices[j]]->location = values[j];

		indices2.clear();
		for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
			if(uniforms_by_index[*j]->size>1)
				indices2.push_back(*j);
		if(!indices2.empty())
		{
			glGetActiveUniformsiv(id, indices2.size(), &indices2[0], GL_UNIFORM_ARRAY_STRIDE, &values[0]);
			for(unsigned j=0; j<indices2.size(); ++j)
				uniforms_by_index[indices2[j]]->array_stride = values[j];
		}

		indices2.clear();
		for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
		{
			GLenum t = uniforms_by_index[*j]->type;
			if(t==GL_FLOAT_MAT4 || t==GL_FLOAT_MAT3 || t==GL_FLOAT_MAT2 ||
				t==GL_FLOAT_MAT2x3 || t==GL_FLOAT_MAT2x4 || t==GL_FLOAT_MAT3x2 ||
				t==GL_FLOAT_MAT3x4 || t==GL_FLOAT_MAT4x2 || t==GL_FLOAT_MAT4x3)
				indices2.push_back(*j);
		}
		if(!indices2.empty())
		{
			glGetActiveUniformsiv(id, indices2.size(), &indices2[0], GL_UNIFORM_MATRIX_STRIDE, &values[0]);
			for(unsigned j=0; j<indices2.size(); ++j)
				uniforms_by_index[indices2[j]]->matrix_stride = values[j];
		}

		sort(info.uniforms.begin(), info.uniforms.end(), uniform_location_compare);
		info.layout_hash = compute_layout_hash(info.uniforms);
		unsigned n_bindings = BufferRange::get_n_uniform_buffer_bindings();
		info.bind_point = info.layout_hash%n_bindings;
		while(used_bind_points.count(info.bind_point))
			info.bind_point = (info.bind_point+1)%n_bindings;
		glUniformBlockBinding(id, i, info.bind_point);
		used_bind_points.insert(info.bind_point);
	}
}

void Program::query_attributes()
{
	unsigned count = get_program_i(id, GL_ACTIVE_ATTRIBUTES);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len = 0;
		int size;
		GLenum type;
		glGetActiveAttrib(id, i, sizeof(name), &len, &size, &type, name);
		if(len && strncmp(name, "gl_", 3))
		{
			if(len>3 && !strcmp(name+len-3, "[0]"))
				name[len-3] = 0;

			AttributeInfo &info = attributes[name];
			info.name = name;
			info.location = glGetAttribLocation(id, name);
			info.size = size;
			info.type = type;
		}
	}
}

Program::LayoutHash Program::compute_layout_hash(const vector<const UniformInfo *> &uniforms)
{
	string layout_descriptor;
	for(vector<const UniformInfo *>::const_iterator i = uniforms.begin(); i!=uniforms.end(); ++i)
		layout_descriptor += format("%d:%s:%x:%d\n", (*i)->location, (*i)->name, (*i)->type, (*i)->size);
	return hash32(layout_descriptor);
}

bool Program::uniform_location_compare(const UniformInfo *uni1, const UniformInfo *uni2)
{
	return uni1->location<uni2->location;
}

string Program::get_info_log() const
{
	GLsizei len = get_program_i(id, GL_INFO_LOG_LENGTH);
	string log(len+1, 0);
	glGetProgramInfoLog(id, len+1, &len, &log[0]);
	log.erase(len);
	return log;
}

const Program::UniformBlockInfo &Program::get_uniform_block_info(const string &name) const
{
	return get_item(uniform_blocks, name);
}

const Program::UniformInfo &Program::get_uniform_info(const string &name) const
{
	return get_item(uniforms, name);
}

int Program::get_uniform_location(const string &n) const
{
	if(n[n.size()-1]==']')
		throw invalid_argument("Program::get_uniform_location");

	UniformMap::const_iterator i = uniforms.find(n);
	if(i==uniforms.end())
		return -1;

	return i->second.block->bind_point<0 ? i->second.location : -1;
}

const Program::AttributeInfo &Program::get_attribute_info(const string &name) const
{
	return get_item(attributes, name);
}

int Program::get_attribute_location(const string &n) const
{
	if(n[n.size()-1]==']')
		throw invalid_argument("Program::get_attribute_location");

	AttributeMap::const_iterator i = attributes.find(n);
	return i!=attributes.end() ? i->second.location : -1;
}

void Program::bind() const
{
	if(!linked)
		throw invalid_operation("Program::bind");

	if(!set_current(this))
		return;

	glUseProgram(id);
}

void Program::unbind()
{
	if(!set_current(0))
		return;

	glUseProgram(0);
}


Program::Loader::Loader(Program &p, Collection &c):
	DataFile::CollectionObjectLoader<Program>(p, &c)
{
	add("module",          &Loader::module);

	// Deprecated
	add("attribute",       &Loader::attribute);
	add("fragment_shader", &Loader::fragment_shader);
	add("geometry_shader", &Loader::geometry_shader);
	add("vertex_shader",   &Loader::vertex_shader);
}

void Program::Loader::finish()
{
	obj.link();
}

void Program::Loader::module(const string &n)
{
	map<string, int> spec_values;
	SpecializationLoader ldr(spec_values);
	load_sub_with(ldr);
	obj.add_stages(get_collection().get<Module>(n), spec_values);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void Program::Loader::attribute(unsigned i, const string &n)
{
	obj.bind_attribute(i, n);
}

void Program::Loader::fragment_shader(const string &src)
{
	obj.attach_shader_owned(new FragmentShader(src));
}

void Program::Loader::geometry_shader(const string &src)
{
	obj.attach_shader_owned(new GeometryShader(src));
}

void Program::Loader::vertex_shader(const string &src)
{
	obj.attach_shader_owned(new VertexShader(src));
}
#pragma GCC diagnostic pop


DataFile::Loader::ActionMap Program::SpecializationLoader::shared_actions;

Program::SpecializationLoader::SpecializationLoader(map<string, int> &sv):
	spec_values(sv)
{
	set_actions(shared_actions);
}

void Program::SpecializationLoader::init_actions()
{
	add("specialize", &SpecializationLoader::specialize_bool);
	add("specialize", &SpecializationLoader::specialize_int);
}

void Program::SpecializationLoader::specialize_bool(const string &name, bool value)
{
	spec_values[name] = value;
}

void Program::SpecializationLoader::specialize_int(const string &name, int value)
{
	spec_values[name] = value;
}

} // namespace GL
} // namespace Msp
