#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include "optimize.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

InlineableFunctionLocator::InlineableFunctionLocator():
	current_function(0),
	return_count(0)
{ }

void InlineableFunctionLocator::visit(FunctionCall &call)
{
	FunctionDeclaration *def = call.declaration;
	if(def)
		def = def->definition;

	if(def)
	{
		unsigned &count = refcounts[def];
		++count;
		/* Don't inline functions which are called more than once or are called
		recursively. */
		if(count>1 || def==current_function)
			inlineable.erase(def);
	}

	TraversingVisitor::visit(call);
}

void InlineableFunctionLocator::visit(FunctionDeclaration &func)
{
	unsigned &count = refcounts[func.definition];
	if(count<=1 && func.parameters.empty())
		inlineable.insert(func.definition);

	SetForScope<FunctionDeclaration *> set(current_function, &func);
	return_count = 0;
	TraversingVisitor::visit(func);
}

void InlineableFunctionLocator::visit(Conditional &cond)
{
	TraversingVisitor::visit(cond);
	inlineable.erase(current_function);
}

void InlineableFunctionLocator::visit(Iteration &iter)
{
	TraversingVisitor::visit(iter);
	inlineable.erase(current_function);
}

void InlineableFunctionLocator::visit(Return &ret)
{
	TraversingVisitor::visit(ret);
	if(return_count)
		inlineable.erase(current_function);
	++return_count;
}


InlineContentInjector::InlineContentInjector():
	source_func(0),
	remap_names(false),
	deps_only(false)
{ }

const string &InlineContentInjector::apply(Stage &stage, FunctionDeclaration &target_func, Block &tgt_blk, const NodeList<Statement>::iterator &ins_pt, FunctionDeclaration &src)
{
	target_block = &tgt_blk;
	source_func = &src;
	for(NodeList<Statement>::iterator i=src.body.body.begin(); i!=src.body.body.end(); ++i)
	{
		r_inlined_statement = 0;
		(*i)->visit(*this);
		if(!r_inlined_statement)
			r_inlined_statement = (*i)->clone();

		SetFlag set_remap(remap_names);
		r_inlined_statement->visit(*this);
		tgt_blk.body.insert(ins_pt, r_inlined_statement);
	}

	NodeReorderer().apply(stage, target_func, dependencies);

	return r_result_name;
}

string InlineContentInjector::create_unused_name(const string &base, bool always_prefix)
{
	string result = base;
	if(always_prefix || target_block->variables.count(result))
		result = format("_%s_%s", source_func->name, base);
	unsigned initial_size = result.size();
	for(unsigned i=1; target_block->variables.count(result); ++i)
	{
		result.erase(initial_size);
		result += format("_%d", i);
	}
	return result;
}

void InlineContentInjector::visit(VariableReference &var)
{
	if(remap_names)
	{
		map<string, VariableDeclaration *>::const_iterator i = variable_map.find(var.name);
		if(i!=variable_map.end())
			var.name = i->second->name;
	}
	else if(var.declaration)
	{
		SetFlag set_deps(deps_only);
		dependencies.insert(var.declaration);
		var.declaration->visit(*this);
	}
}

void InlineContentInjector::visit(InterfaceBlockReference &iface)
{
	if(!remap_names && iface.declaration)
	{
		SetFlag set_deps(deps_only);
		dependencies.insert(iface.declaration);
		iface.declaration->visit(*this);
	}
}

void InlineContentInjector::visit(FunctionCall &call)
{
	if(!remap_names && call.declaration)
		dependencies.insert(call.declaration);
	TraversingVisitor::visit(call);
}

void InlineContentInjector::visit(VariableDeclaration &var)
{
	TraversingVisitor::visit(var);

	if(var.type_declaration)
	{
		SetFlag set_deps(deps_only);
		dependencies.insert(var.type_declaration);
		var.type_declaration->visit(*this);
	}

	if(!remap_names && !deps_only)
	{
		RefPtr<VariableDeclaration> inlined_var = var.clone();
		inlined_var->name = create_unused_name(var.name, false);
		r_inlined_statement = inlined_var;

		variable_map[var.name] = inlined_var.get();
	}
}

void InlineContentInjector::visit(Return &ret)
{
	TraversingVisitor::visit(ret);

	if(ret.expression)
	{
		/* Create a new variable to hold the return value of the inlined
		function. */
		r_result_name = create_unused_name("return", true);
		RefPtr<VariableDeclaration> var = new VariableDeclaration;
		var->source = ret.source;
		var->line = ret.line;
		var->type = source_func->return_type;
		var->name = r_result_name;
		var->init_expression = ret.expression->clone();
		r_inlined_statement = var;
	}
}


FunctionInliner::FunctionInliner():
	current_function(0),
	r_any_inlined(false)
{ }

bool FunctionInliner::apply(Stage &s)
{
	stage = &s;
	inlineable = InlineableFunctionLocator().apply(s);
	r_any_inlined = false;
	s.content.visit(*this);
	return r_any_inlined;
}

void FunctionInliner::visit(RefPtr<Expression> &ptr)
{
	r_inline_result = 0;
	ptr->visit(*this);
	if(r_inline_result)
	{
		ptr = r_inline_result;
		r_any_inlined = true;
	}
	r_inline_result = 0;
}

void FunctionInliner::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	SetForScope<NodeList<Statement>::iterator> save_insert_point(insert_point, block.body.begin());
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		insert_point = i;
		(*i)->visit(*this);
	}
}

void FunctionInliner::visit(FunctionCall &call)
{
	for(NodeArray<Expression>::iterator i=call.arguments.begin(); i!=call.arguments.end(); ++i)
		visit(*i);

	FunctionDeclaration *def = call.declaration;
	if(def)
		def = def->definition;

	if(def && inlineable.count(def))
	{
		string result_name = InlineContentInjector().apply(*stage, *current_function, *current_block, insert_point, *def);

		// This will later get removed by UnusedVariableRemover.
		if(result_name.empty())
			result_name = "msp_unused_from_inline";

		RefPtr<VariableReference> ref = new VariableReference;
		ref->name = result_name;
		r_inline_result = ref;

		/* Inlined variables need to be resolved before this function can be
		inlined further. */
		inlineable.erase(current_function);
	}
}

void FunctionInliner::visit(FunctionDeclaration &func)
{
	SetForScope<FunctionDeclaration *> set_func(current_function, &func);
	TraversingVisitor::visit(func);
}

void FunctionInliner::visit(Iteration &iter)
{
	/* Visit the initialization statement before entering the loop body so the
	inlined statements get inserted outside. */
	if(iter.init_statement)
		iter.init_statement->visit(*this);

	SetForScope<Block *> set_block(current_block, &iter.body);
	/* Skip the condition and loop expression parts because they're not properly
	inside the body block.  Inlining anything into them will require a more
	comprehensive transformation. */
	iter.body.visit(*this);
}


ExpressionInliner::ExpressionInfo::ExpressionInfo():
	expression(0),
	assign_scope(0),
	inline_point(0),
	inner_oper(0),
	outer_oper(0),
	inline_on_rhs(false),
	trivial(false),
	available(true)
{ }


ExpressionInliner::ExpressionInliner():
	r_ref_info(0),
	r_any_inlined(false),
	r_trivial(false),
	mutating(false),
	iteration_init(false),
	iteration_body(0),
	r_oper(0)
{ }

bool ExpressionInliner::apply(Stage &s)
{
	s.content.visit(*this);
	return r_any_inlined;
}

void ExpressionInliner::visit_and_record(RefPtr<Expression> &ptr, const Operator *outer_oper, bool on_rhs)
{
	r_ref_info = 0;
	ptr->visit(*this);
	if(r_ref_info && r_ref_info->expression && r_ref_info->available)
	{
		if(iteration_body && !r_ref_info->trivial)
		{
			/* Don't inline non-trivial expressions which were assigned outside
			an iteration statement.  The iteration may run multiple times, which
			would cause the expression to also be evaluated multiple times. */
			Block *i = r_ref_info->assign_scope;
			for(; (i && i!=iteration_body); i=i->parent) ;
			if(!i)
				return;
		}

		r_ref_info->outer_oper = outer_oper;
		if(r_ref_info->trivial)
			inline_expression(*r_ref_info->expression, ptr, outer_oper, r_ref_info->inner_oper, on_rhs);
		else
		{
			/* Record the inline point for a non-trivial expression but don't
			inline it yet.  It might turn out it shouldn't be inlined after all. */
			r_ref_info->inline_point = &ptr;
			r_ref_info->inline_on_rhs = on_rhs;
		}
	}
	r_ref_info = 0;
}

void ExpressionInliner::inline_expression(Expression &expr, RefPtr<Expression> &ptr, const Operator *outer_oper, const Operator *inner_oper, bool on_rhs)
{
	unsigned outer_precedence = (outer_oper ? outer_oper->precedence : 20);
	unsigned inner_precedence = (inner_oper ? inner_oper->precedence : 0);

	bool needs_parentheses = (inner_precedence>=outer_precedence);
	if(inner_oper && inner_oper==outer_oper)
		// Omit parentheses if the operator's natural grouping works out.
		needs_parentheses = (inner_oper->assoc!=Operator::ASSOCIATIVE && on_rhs!=(inner_oper->assoc==Operator::RIGHT_TO_LEFT));

	if(needs_parentheses)
	{
		RefPtr<ParenthesizedExpression> parexpr = new ParenthesizedExpression;
		parexpr->expression = expr.clone();
		ptr = parexpr;
	}
	else
		ptr = expr.clone();

	r_any_inlined = true;
}

void ExpressionInliner::visit(Block &block)
{
	TraversingVisitor::visit(block);

	for(map<string, VariableDeclaration *>::iterator i=block.variables.begin(); i!=block.variables.end(); ++i)
	{
		map<Assignment::Target, ExpressionInfo>::iterator j = expressions.lower_bound(i->second);
		for(; (j!=expressions.end() && j->first.declaration==i->second); )
		{
			if(j->second.expression && j->second.inline_point)
				inline_expression(*j->second.expression, *j->second.inline_point, j->second.outer_oper, j->second.inner_oper, j->second.inline_on_rhs);

			expressions.erase(j++);
		}
	}

	/* Expressions assigned in this block may depend on local variables of the
	block.  If this is a conditionally executed block, the assignments might not
	always happen.  Mark the expressions as not available to any outer blocks. */
	for(map<Assignment::Target, ExpressionInfo>::iterator i=expressions.begin(); i!=expressions.end(); ++i)
		if(i->second.assign_scope==&block)
			i->second.available = false;
}

void ExpressionInliner::visit(RefPtr<Expression> &expr)
{
	visit_and_record(expr, 0, false);
}

void ExpressionInliner::visit(VariableReference &var)
{
	if(var.declaration)
	{
		map<Assignment::Target, ExpressionInfo>::iterator i = expressions.find(var.declaration);
		if(i!=expressions.end())
		{
			/* If a non-trivial expression is referenced multiple times, don't
			inline it. */
			if(i->second.inline_point && !i->second.trivial)
				i->second.expression = 0;
			/* Mutating expressions are analogous to self-referencing assignments
			and prevent inlining. */
			if(mutating)
				i->second.expression = 0;
			r_ref_info = &i->second;
		}
	}
}

void ExpressionInliner::visit(MemberAccess &memacc)
{
	visit_and_record(memacc.left, memacc.oper, false);
	r_oper = memacc.oper;
	r_trivial = false;
}

void ExpressionInliner::visit(Swizzle &swizzle)
{
	visit_and_record(swizzle.left, swizzle.oper, false);
	r_oper = swizzle.oper;
	r_trivial = false;
}

void ExpressionInliner::visit(UnaryExpression &unary)
{
	SetFlag set_target(mutating, mutating || unary.oper->token[1]=='+' || unary.oper->token[1]=='-');
	visit_and_record(unary.expression, unary.oper, false);
	r_oper = unary.oper;
	r_trivial = false;
}

void ExpressionInliner::visit(BinaryExpression &binary)
{
	visit_and_record(binary.left, binary.oper, false);
	{
		SetFlag clear_target(mutating, false);
		visit_and_record(binary.right, binary.oper, true);
	}
	r_oper = binary.oper;
	r_trivial = false;
}

void ExpressionInliner::visit(Assignment &assign)
{
	{
		SetFlag set_target(mutating);
		visit_and_record(assign.left, assign.oper, false);
	}
	r_oper = 0;
	visit_and_record(assign.right, assign.oper, true);

	map<Assignment::Target, ExpressionInfo>::iterator i = expressions.find(assign.target);
	if(i!=expressions.end())
	{
		/* Self-referencing assignments can't be inlined without additional
		work.  Just clear any previous expression. */
		i->second.expression = (assign.self_referencing ? 0 : assign.right.get());
		i->second.assign_scope = current_block;
		i->second.inline_point = 0;
		i->second.inner_oper = r_oper;
		i->second.available = true;
	}

	r_oper = assign.oper;
	r_trivial = false;
}

void ExpressionInliner::visit(TernaryExpression &ternary)
{
	visit_and_record(ternary.condition, ternary.oper, false);
	visit_and_record(ternary.true_expr, ternary.oper, false);
	visit_and_record(ternary.false_expr, ternary.oper, true);
	r_oper = ternary.oper;
	r_trivial = false;
}

void ExpressionInliner::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);
	r_oper = 0;
	r_trivial = false;
}

void ExpressionInliner::visit(VariableDeclaration &var)
{
	r_oper = 0;
	r_trivial = true;
	TraversingVisitor::visit(var);

	bool constant = var.constant;
	if(constant && var.layout)
	{
		for(vector<Layout::Qualifier>::const_iterator i=var.layout->qualifiers.begin(); (constant && i!=var.layout->qualifiers.end()); ++i)
			constant = (i->name!="constant_id");
	}

	/* Only inline global variables if they're constant and have trivial
	initializers.  Non-constant variables could change in ways which are hard to
	analyze and non-trivial expressions could be expensive to inline.  */
	if((current_block->parent || (constant && r_trivial)) && var.interface.empty())
	{
		ExpressionInfo &info = expressions[&var];
		/* Assume variables declared in an iteration initialization statement
		will have their values change throughout the iteration. */
		info.expression = (iteration_init ? 0 : var.init_expression.get());
		info.assign_scope = current_block;
		info.inner_oper = r_oper;
		info.trivial = r_trivial;
	}
}

void ExpressionInliner::visit(Iteration &iter)
{
	SetForScope<Block *> set_block(current_block, &iter.body);
	if(iter.init_statement)
	{
		SetFlag set_init(iteration_init);
		iter.init_statement->visit(*this);
	}

	SetForScope<Block *> set_body(iteration_body, &iter.body);
	if(iter.condition)
		visit(iter.condition);
	iter.body.visit(*this);
	if(iter.loop_expression)
		visit(iter.loop_expression);
}


void ConstantConditionEliminator::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, nodes_to_remove);
}

void ConstantConditionEliminator::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		insert_point = i;
		(*i)->visit(*this);
	}
}

void ConstantConditionEliminator::visit(Conditional &cond)
{
	if(Literal *literal = dynamic_cast<Literal *>(cond.condition.get()))
		if(literal->value.check_type<bool>())
		{
			Block &block = (literal->value.value<bool>() ? cond.body : cond.else_body);
			current_block->body.splice(insert_point, block.body);
			nodes_to_remove.insert(&cond);
			return;
		}

	TraversingVisitor::visit(cond);
}

void ConstantConditionEliminator::visit(Iteration &iter)
{
	if(iter.condition)
	{
		/* If the loop condition is always false on the first iteration, the
		entire loop can be removed */
		ExpressionEvaluator::ValueMap values;
		if(VariableDeclaration *var = dynamic_cast<VariableDeclaration *>(iter.init_statement.get()))
			values[var] = var->init_expression.get();
		ExpressionEvaluator eval(values);
		iter.condition->visit(eval);
		if(eval.is_result_valid() && !eval.get_result())
		{
			nodes_to_remove.insert(&iter);
			return;
		}
	}

	TraversingVisitor::visit(iter);
}


bool UnusedTypeRemover::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, unused_nodes);
	return !unused_nodes.empty();
}

void UnusedTypeRemover::visit(Literal &literal)
{
	unused_nodes.erase(literal.type);
}

void UnusedTypeRemover::visit(UnaryExpression &unary)
{
	unused_nodes.erase(unary.type);
	TraversingVisitor::visit(unary);
}

void UnusedTypeRemover::visit(BinaryExpression &binary)
{
	unused_nodes.erase(binary.type);
	TraversingVisitor::visit(binary);
}

void UnusedTypeRemover::visit(TernaryExpression &ternary)
{
	unused_nodes.erase(ternary.type);
	TraversingVisitor::visit(ternary);
}

void UnusedTypeRemover::visit(FunctionCall &call)
{
	unused_nodes.erase(call.type);
	TraversingVisitor::visit(call);
}

void UnusedTypeRemover::visit(BasicTypeDeclaration &type)
{
	if(type.base_type)
		unused_nodes.erase(type.base_type);
	unused_nodes.insert(&type);
}

void UnusedTypeRemover::visit(ImageTypeDeclaration &type)
{
	if(type.base_type)
		unused_nodes.erase(type.base_type);
	unused_nodes.insert(&type);
}

void UnusedTypeRemover::visit(StructDeclaration &strct)
{
	unused_nodes.insert(&strct);
	TraversingVisitor::visit(strct);
}

void UnusedTypeRemover::visit(VariableDeclaration &var)
{
	unused_nodes.erase(var.type_declaration);
}

void UnusedTypeRemover::visit(InterfaceBlock &iface)
{
	unused_nodes.erase(iface.type_declaration);
}

void UnusedTypeRemover::visit(FunctionDeclaration &func)
{
	unused_nodes.erase(func.return_type_declaration);
	TraversingVisitor::visit(func);
}


UnusedVariableRemover::UnusedVariableRemover():
	stage(0),
	interface_block(0),
	r_assignment(0),
	assignment_target(false),
	r_side_effects(false)
{ }

bool UnusedVariableRemover::apply(Stage &s)
{
	stage = &s;
	s.content.visit(*this);

	for(list<AssignmentInfo>::const_iterator i=assignments.begin(); i!=assignments.end(); ++i)
		if(i->used_by.empty())
			unused_nodes.insert(i->node);

	for(map<string, InterfaceBlock *>::const_iterator i=s.interface_blocks.begin(); i!=s.interface_blocks.end(); ++i)
		if(i->second->instance_name.empty())
			unused_nodes.insert(i->second);

	for(BlockVariableMap::const_iterator i=variables.begin(); i!=variables.end(); ++i)
	{
		if(i->second.output)
		{
			/* The last visible assignments of output variables are used by the
			next stage or the API. */
			for(vector<AssignmentInfo *>::const_iterator j=i->second.assignments.begin(); j!=i->second.assignments.end(); ++j)
				unused_nodes.erase((*j)->node);
		}

		if(!i->second.output && !i->second.referenced)
		{
			// Don't remove variables from inside interface blocks.
			if(!i->second.interface_block)
				unused_nodes.insert(i->first);
		}
		else if(i->second.interface_block)
			// Interface blocks are kept if even one member is used.
			unused_nodes.erase(i->second.interface_block);
	}

	NodeRemover().apply(s, unused_nodes);

	return !unused_nodes.empty();
}

void UnusedVariableRemover::referenced(const Assignment::Target &target, Node &node)
{
	VariableInfo &var_info = variables[target.declaration];
	var_info.referenced = true;
	if(!assignment_target)
	{
		for(vector<AssignmentInfo *>::const_iterator i=var_info.assignments.begin(); i!=var_info.assignments.end(); ++i)
			(*i)->used_by.push_back(&node);
	}
}

void UnusedVariableRemover::visit(VariableReference &var)
{
	referenced(var.declaration, var);
}

void UnusedVariableRemover::visit(InterfaceBlockReference &iface)
{
	referenced(iface.declaration, iface);
}

void UnusedVariableRemover::visit(UnaryExpression &unary)
{
	TraversingVisitor::visit(unary);
	if(unary.oper->token[1]=='+' || unary.oper->token[1]=='-')
		r_side_effects = true;
}

void UnusedVariableRemover::visit(BinaryExpression &binary)
{
	if(binary.oper->token[0]=='[')
	{
		binary.left->visit(*this);
		SetFlag set(assignment_target, false);
		binary.right->visit(*this);
	}
	else
		TraversingVisitor::visit(binary);
}

void UnusedVariableRemover::visit(Assignment &assign)
{
	{
		SetFlag set(assignment_target, (assign.oper->token[0]=='='));
		assign.left->visit(*this);
	}
	assign.right->visit(*this);
	r_assignment = &assign;
	r_side_effects = true;
}

void UnusedVariableRemover::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);
	/* Treat function calls as having side effects so expression statements
	consisting of nothing but a function call won't be optimized away. */
	r_side_effects = true;

	if(stage->type==Stage::GEOMETRY && call.name=="EmitVertex")
	{
		for(map<Statement *, VariableInfo>::const_iterator i=variables.begin(); i!=variables.end(); ++i)
			if(i->second.output)
				referenced(i->first, call);
	}
}

void UnusedVariableRemover::record_assignment(const Assignment::Target &target, Node &node)
{
	assignments.push_back(AssignmentInfo());
	AssignmentInfo &assign_info = assignments.back();
	assign_info.node = &node;
	assign_info.target = target;

	/* An assignment to the target hides any assignments to the same target or
	its subfields. */
	VariableInfo &var_info = variables[target.declaration];
	for(unsigned i=0; i<var_info.assignments.size(); ++i)
	{
		const Assignment::Target &t = var_info.assignments[i]->target;

		bool subfield = (t.chain_len>=target.chain_len);
		for(unsigned j=0; (subfield && j<target.chain_len); ++j)
			subfield = (t.chain[j]==target.chain[j]);

		if(subfield)
			var_info.assignments.erase(var_info.assignments.begin()+i);
		else
			++i;
	}

	var_info.assignments.push_back(&assign_info);
}

void UnusedVariableRemover::visit(ExpressionStatement &expr)
{
	r_assignment = 0;
	r_side_effects = false;
	TraversingVisitor::visit(expr);
	if(r_assignment && r_assignment->target.declaration)
		record_assignment(r_assignment->target, expr);
	if(!r_side_effects)
		unused_nodes.insert(&expr);
}

void UnusedVariableRemover::visit(VariableDeclaration &var)
{
	VariableInfo &var_info = variables[&var];
	var_info.interface_block = interface_block;

	/* Mark variables as output if they're used by the next stage or the
	graphics API. */
	if(interface_block)
		var_info.output = (interface_block->interface=="out" && (interface_block->linked_block || !interface_block->name.compare(0, 3, "gl_")));
	else
		var_info.output = (var.interface=="out" && (stage->type==Stage::FRAGMENT || var.linked_declaration || !var.name.compare(0, 3, "gl_")));

	if(var.init_expression)
	{
		var_info.initialized = true;
		record_assignment(&var, *var.init_expression);
	}
	TraversingVisitor::visit(var);
}

void UnusedVariableRemover::visit(InterfaceBlock &iface)
{
	if(iface.instance_name.empty())
	{
		SetForScope<InterfaceBlock *> set_block(interface_block, &iface);
		iface.struct_declaration->members.visit(*this);
	}
	else
	{
		VariableInfo &var_info = variables[&iface];
		var_info.output = (iface.interface=="out" && (iface.linked_block || !iface.name.compare(0, 3, "gl_")));
	}
}

void UnusedVariableRemover::merge_variables(const BlockVariableMap &other_vars)
{
	for(BlockVariableMap::const_iterator i=other_vars.begin(); i!=other_vars.end(); ++i)
	{
		BlockVariableMap::iterator j = variables.find(i->first);
		if(j!=variables.end())
		{
			/* The merged blocks started as copies of each other so any common
			assignments must be in the beginning. */
			unsigned k = 0;
			for(; (k<i->second.assignments.size() && k<j->second.assignments.size()); ++k)
				if(i->second.assignments[k]!=j->second.assignments[k])
					break;

			// Remaining assignments are unique to each block; merge them.
			j->second.assignments.insert(j->second.assignments.end(), i->second.assignments.begin()+k, i->second.assignments.end());
			j->second.referenced |= i->second.referenced;
		}
		else
			variables.insert(*i);
	}
}

void UnusedVariableRemover::visit(FunctionDeclaration &func)
{
	if(func.body.body.empty())
		return;

	BlockVariableMap saved_vars = variables;
	// Assignments from other functions should not be visible.
	for(BlockVariableMap::iterator i=variables.begin(); i!=variables.end(); ++i)
		i->second.assignments.resize(i->second.initialized);
	TraversingVisitor::visit(func);
	swap(variables, saved_vars);
	merge_variables(saved_vars);

	/* Always treat function parameters as referenced.  Removing unused
	parameters is not currently supported. */
	for(NodeArray<VariableDeclaration>::iterator i=func.parameters.begin(); i!=func.parameters.end(); ++i)
	{
		BlockVariableMap::iterator j = variables.find(i->get());
		if(j!=variables.end())
			j->second.referenced = true;
	}
}

void UnusedVariableRemover::visit(Conditional &cond)
{
	cond.condition->visit(*this);
	BlockVariableMap saved_vars = variables;
	cond.body.visit(*this);
	swap(saved_vars, variables);
	cond.else_body.visit(*this);

	/* Visible assignments after the conditional is the union of those visible
	at the end of the if and else blocks.  If there was no else block, then it's
	the union of the if block and the state before it. */
	merge_variables(saved_vars);
}

void UnusedVariableRemover::visit(Iteration &iter)
{
	BlockVariableMap saved_vars = variables;
	TraversingVisitor::visit(iter);

	/* Merge assignments from the iteration, without clearing previous state.
	Further analysis is needed to determine which parts of the iteration body
	are always executed, if any. */
	merge_variables(saved_vars);
}


bool UnusedFunctionRemover::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, unused_nodes);
	return !unused_nodes.empty();
}

void UnusedFunctionRemover::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);

	unused_nodes.erase(call.declaration);
	if(call.declaration && call.declaration->definition!=call.declaration)
		used_definitions.insert(call.declaration->definition);
}

void UnusedFunctionRemover::visit(FunctionDeclaration &func)
{
	TraversingVisitor::visit(func);

	if((func.name!="main" || func.body.body.empty()) && !used_definitions.count(&func))
		unused_nodes.insert(&func);
}

} // namespace SL
} // namespace GL
} // namespace Msp
