#ifndef MSP_GL_SL_VISITOR_H_
#define MSP_GL_SL_VISITOR_H_

#include <set>
#include <vector>
#include "syntax.h"

namespace Msp {
namespace GL {
namespace SL {

/** Base class for all node visitors. */
class NodeVisitor
{
protected:
	NodeVisitor() { }
public:
	virtual ~NodeVisitor() { }

	virtual void visit(Block &) { }
	virtual void visit(Literal &) { }
	virtual void visit(ParenthesizedExpression &) { }
	virtual void visit(VariableReference &) { }
	virtual void visit(InterfaceBlockReference &) { }
	virtual void visit(MemberAccess &) { }
	virtual void visit(Swizzle &) { }
	virtual void visit(UnaryExpression &) { }
	virtual void visit(BinaryExpression &) { }
	virtual void visit(Assignment &) { }
	virtual void visit(TernaryExpression &) { }
	virtual void visit(FunctionCall &) { }
	virtual void visit(ExpressionStatement &) { }
	virtual void visit(Import &) { }
	virtual void visit(Precision &) { }
	virtual void visit(Layout &) { }
	virtual void visit(InterfaceLayout &) { }
	virtual void visit(BasicTypeDeclaration &) { }
	virtual void visit(ImageTypeDeclaration &) { }
	virtual void visit(StructDeclaration &) { }
	virtual void visit(VariableDeclaration &) { }
	virtual void visit(InterfaceBlock &) { }
	virtual void visit(FunctionDeclaration &) { }
	virtual void visit(Conditional &) { }
	virtual void visit(Iteration &) { }
	virtual void visit(Passthrough &) { }
	virtual void visit(Return &) { }
	virtual void visit(Jump &) { }
};

/** An intermediate base visitor class which traverses the syntax tree. */
class TraversingVisitor: public NodeVisitor
{
protected:
	Block *current_block;

	TraversingVisitor(): current_block(0) { }

public:
	virtual void enter(Block &) { }
	virtual void visit(Block &);
	virtual void visit(RefPtr<Expression> &);
	virtual void visit(ParenthesizedExpression &);
	virtual void visit(MemberAccess &);
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	virtual void visit(FunctionCall &);
	virtual void visit(ExpressionStatement &);
	virtual void visit(InterfaceLayout &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Conditional &);
	virtual void visit(Iteration &);
	virtual void visit(Passthrough &);
	virtual void visit(Return &);
};

/** Removes a set of nodes from the syntax tree. */
class NodeRemover: private TraversingVisitor
{
private:
	Stage *stage;
	const std::set<Node *> *to_remove;
	bool recursive_remove;

public:
	NodeRemover();

	void apply(Stage &, const std::set<Node *> &);

private:
	template<typename T>
	void remove_from_map(std::map<std::string, T *> &, const std::string &, T &);

	virtual void visit(Block &);
	void visit(TypeDeclaration &);
	virtual void visit(BasicTypeDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(ImageTypeDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(StructDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Iteration &);
};

/** Reorders a set of nodes so they appear before another node.  Only nodes
on the same hierarchly level as the target node are reordered. */
class NodeReorderer: private TraversingVisitor
{
private:
	Node *reorder_before;
	const std::set<Node *> *to_reorder;

public:
	NodeReorderer();

	void apply(Stage &, Node &, const std::set<Node *> &);

private:
	virtual void visit(Block &);
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
